class TemplatecoException(BaseException):
    """
    Generic Templateco exception.
    """

    pass


class IncompatiblePluginException(TemplatecoException):
    """
    Indicates that a Templateco plugin has been loaded, but it is not compatible.
    """

    pass


class NoPluginsFoundException(TemplatecoException):
    """
    Called when Templateco has been invoked, but no plugins are specified.
    """

    pass


class UnknownOperatingModeException(TemplatecoException):
    """
    Called when Templateco is missing the operating mode e.g. prefixes or plugins.
    """

    pass


class PluginOperationException(TemplatecoException):
    """
    Called during a Templateco plugin's operation to be echoed back to the user.

    This should be called by Plugin authors when they wish to pass a clear message back
    up to the user that _something_ they have selected isn't correct.
    """

    pass



from .exceptions import (
    IncompatiblePluginException,
    NoPluginsFoundException,
    PluginOperationException,
    TemplatecoException,
    UnknownOperatingModeException,
)
from .templateco import templateco
from .typing import (
    folders_to_template_decorator,
    get_questions_decorator,
    post_template_hook_decorator,
    pre_template_hook_decorator,
)

__all__ = [
    "templateco",
    "IncompatiblePluginException",
    "NoPluginsFoundException",
    "PluginOperationException",
    "TemplatecoException",
    "UnknownOperatingModeException",
    "get_questions_decorator",
    "pre_template_hook_decorator",
    "folders_to_template_decorator",
    "post_template_hook_decorator",
]



import logging
from dataclasses import dataclass
from functools import lru_cache
from importlib import import_module
from pathlib import Path
from typing import Generator, Iterable, List, Optional

from importlib_metadata import Distribution, distributions

from templateco.exceptions import IncompatiblePluginException
from templateco.typing import (
    FoldersToTemplate,
    GetQuestions,
    PostTemplateHook,
    PreTemplateHook,
)

logger = logging.getLogger(__name__)


@dataclass
class TemplatecoPlugin:
    """
    Templateco Plugin.
    """

    friendly_name: str
    path_to_module: Path
    compatible_templates: List[str]
    selected_by_default: bool
    get_questions: GetQuestions
    folders_to_template: FoldersToTemplate
    pre_template_hook: Optional[PreTemplateHook]
    post_template_hook: Optional[PostTemplateHook]


@lru_cache(maxsize=None)
def import_templateco_plugin(name: str) -> TemplatecoPlugin:
    """
    Import a compatible Templateco package, and return a TemplatecoPlugin.

    :param name: Name of Python package.
    :return: Templateco Plugin.
    """
    module = import_module(name.replace("-", "_"))

    try:
        return TemplatecoPlugin(
            friendly_name=module.__dict__["friendly_name"],
            path_to_module=Path(str(module.__file__)).parent,
            compatible_templates=module.__dict__["compatible_templates"],
            selected_by_default=module.__dict__.get("selected_by_default", False),
            get_questions=module.__dict__["get_questions"],
            folders_to_template=module.__dict__["folders_to_template"],
            pre_template_hook=module.__dict__.get("pre_template_hook", None),
            post_template_hook=module.__dict__.get("post_template_hook", None),
        )
    except KeyError as err:
        logger.debug(f"[{name}] Failed to parse plugin information.")
        logger.debug(err)
        raise IncompatiblePluginException(err)


def list_compatible_plugins(
    namespaces: List[str],
) -> Generator[TemplatecoPlugin, None, None]:
    """
    Return a list of compatible Templateco plugins.

    :param namespaces: List of Templateco plugin namespaces to accept.
    :return: List of Templateco plugins.
    """
    logger.debug("Getting list of Python distributions.")

    installed_distributions: Iterable[Distribution] = distributions()  # type: ignore

    for distribution in installed_distributions:
        if "-" in distribution.name:
            split_package = distribution.name.split("-")
            if split_package[0] in namespaces and distribution.name[-6:] == "plugin":
                logger.debug(f"[{distribution.name}] Found potential plugin.")
                try:
                    yield import_templateco_plugin(distribution.name)
                    logger.debug(f"[{distribution.name}] Successfully loaded plugin.")
                except IncompatiblePluginException:
                    logger.debug(f"[{distribution.name}] Plugin incompatible.")



import logging
from pathlib import Path
from typing import List

from colorama import Fore, Style, init
from questionary import Choice, checkbox, unsafe_prompt

from templateco.exceptions import NoPluginsFoundException, PluginOperationException
from templateco.output import generate_output
from templateco.plugin import TemplatecoPlugin, list_compatible_plugins

logger = logging.getLogger(__name__)


def _execute_plugin(
    template: str, output_folder: str, plugin: TemplatecoPlugin
) -> None:
    """
    Execute a single Templateco plugin.

    This function prompts for the plugin's questions, runs the hooks and generates
    the template files.

    :param template:
    :param output_folder: Folder to output to.
    :param plugin: Plugin to execute.
    :return: None.
    """
    logger.debug(f"[{plugin.friendly_name}] Running plugin.")
    try:
        folder_name = Path(output_folder).name
        logger.debug(f"[{plugin.friendly_name}] Prompting for interaction.")

        data = unsafe_prompt(
            plugin.get_questions(template=template, folder_name=folder_name)
        )

        if plugin.pre_template_hook:
            logger.debug(f"[{plugin.friendly_name}] Running pre-template hook.")
            plugin.pre_template_hook(
                template=template, folder_name=folder_name, data=data
            )

        logger.debug(f"[{plugin.friendly_name}] Determining template from answers.")
        folders = plugin.folders_to_template(
            template=template, folder_name=folder_name, data=data
        )

        logger.debug(f"[{plugin.friendly_name}] Templating folder structure.")
        generate_output(plugin.path_to_module, Path() / output_folder, folders, data)
        logger.debug(f"[{plugin.friendly_name}] Successfully templated.")

        if plugin.post_template_hook:
            logger.debug(f"[{plugin.friendly_name}] Running post-template hook.")
            plugin.post_template_hook(
                template=template,
                folder_name=folder_name,
                output_folder=Path() / output_folder,
                data=data,
            )
    except PluginOperationException:
        logger.error(f"[{plugin.friendly_name}] Plugin raised Operation Exception.")
        logger.debug(
            f"[{plugin.friendly_name}] These exceptions should be raised to the user."
        )
        raise
    except Exception as err:
        logger.error(f"[{plugin.friendly_name}] Plugin raised an unhandled exception.")
        logger.error(err)


def templateco(
    template: str,
    output_folder: str,
    namespaces: List[str] = ["templateco"],
) -> None:
    """
    Execute Templateco, and produce a templated folder.

    :param template: Name of template to load plugins for.
    :param output_folder: Folder to generate e.g. $(PWD)/output_folder.
    :param namespaces: List of Templateco plugin namespaces to accept.
    :return: None.
    """
    # Initialise Colorama.
    init(autoreset=True)
    print(
        Fore.GREEN
        + Style.BRIGHT
        + "[ Templateco ]"
        + Style.RESET_ALL
        + " - A Templating Ecosystem"
    )

    logger.debug("Instantiating Templateco.")

    # Get a complete list of all compatible Templateco plugins.
    plugins = [
        plugin
        for plugin in list_compatible_plugins(namespaces)
        if template in plugin.compatible_templates
    ]

    logger.debug(f"Found {len(plugins)} plugin(s).")

    if len(plugins) == 0:
        logger.debug("Query resulted in zero plugins, raising NoPlugins exception.")
        raise NoPluginsFoundException()
    elif len(plugins) > 1:
        logger.debug("Multiple plugins found, asking user for choices.")
        # Ask the user which plugins should be executed.
        plugins = checkbox(
            "Select plugins to use",
            choices=[
                Choice(
                    plugin.friendly_name,
                    checked=plugin.selected_by_default,
                    value=plugin,
                )
                for plugin in plugins
            ],
        ).unsafe_ask()

    for plugin in plugins:
        print("\nConfiguring " + Fore.GREEN + Style.BRIGHT + f"{plugin.friendly_name}")
        _execute_plugin(template, output_folder, plugin)

    print("\nSuccessfully templated " + Style.BRIGHT + Fore.GREEN + output_folder)



from pathlib import Path
from typing import Any, Dict, Iterable, List, Mapping

from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class GetQuestions(Protocol):
    """
    The get_questions typing definition.
    """

    def __call__(
        self, *, template: str, folder_name: str
    ) -> Iterable[Mapping[str, Any]]:
        """
        Return a Questionary dictionary-type iterable.

        :param template: Name of the requested template.
        :param folder_name: Name of the top-level folder.
        :return: Dictionary of data passed to the template.
        """
        pass


@runtime_checkable
class FoldersToTemplate(Protocol):
    """
    The folders_to_template typing definition.
    """

    def __call__(
        self, *, template: str, folder_name: str, data: Dict[str, Any]
    ) -> List[str]:
        """
        Take all input and returns a list of folders within the plugin.

        These folders are used to specify which should be templated, and intended to
        support optional configuration.

        :param template: Name of the requested template.
        :param folder_name: Name of the top-level folder.
        :param data: Dictionary of data passed to the template.
        :return: List of folders to template.
        """
        pass


@runtime_checkable
class PreTemplateHook(Protocol):
    """
    The pre_template_hook typing definition.
    """

    def __call__(
        self, *, template: str, folder_name: str, data: Dict[str, Any]
    ) -> None:
        """
        Run before a plugin creates a template.

        This optional function is intended to be used to modify the data dictionary
        that's passed to Jinja2 for the templating.

        :param template: Name of the requested template.
        :param folder_name: Name of the top-level folder.
        :param data: Dictionary of data passed to the template.
        :return: Nothing
        """
        pass


@runtime_checkable
class PostTemplateHook(Protocol):
    """
    The post_template_hook typing definition.
    """

    def __call__(
        self,
        *,
        template: str,
        folder_name: str,
        output_folder: Path,
        data: Dict[str, Any]
    ) -> None:
        """
        Run after a plugin has successfully templated.

        This optional function is intended to be used to run steps after the templating
        has been run, and passes in the path to the final directory. This can be used
        to run CLI commands, or alter the folder data if wished.

        :param template: Name of the requested template.
        :param folder_name: Name of the top-level folder.
        :param output_folder: Path to the generated folder.
        :param data: Dictionary of data passed to the template.
        :return: Nothing
        """
        pass


def get_questions_decorator(func: GetQuestions) -> GetQuestions:
    """
    Decorate the get_questions plugin function.
    """
    return func


def folders_to_template_decorator(func: FoldersToTemplate) -> FoldersToTemplate:
    """
    Decorate the folders_to_template plugin function.
    """
    return func


def pre_template_hook_decorator(func: PreTemplateHook) -> PreTemplateHook:
    """
    Decorate the pre_template_hook plugin function.
    """
    return func


def post_template_hook_decorator(func: PostTemplateHook) -> PostTemplateHook:
    """
    Decorate the post_template_hook plugin function.
    """
    return func



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '..'}

packages = \
['templateco', 'templateco.dist.templateco-0.1.2']

package_data = \
{'': ['*'],
 'templateco': ['.mypy_cache/*',
                '.mypy_cache/3.7/*',
                '.mypy_cache/3.7/_pytest/*',
                '.mypy_cache/3.7/_pytest/_code/*',
                '.mypy_cache/3.7/_pytest/_io/*',
                '.mypy_cache/3.7/_pytest/assertion/*',
                '.mypy_cache/3.7/_pytest/config/*',
                '.mypy_cache/3.7/_pytest/mark/*',
                '.mypy_cache/3.7/_typeshed/*',
                '.mypy_cache/3.7/asyncio/*',
                '.mypy_cache/3.7/attr/*',
                '.mypy_cache/3.7/collections/*',
                '.mypy_cache/3.7/concurrent/*',
                '.mypy_cache/3.7/concurrent/futures/*',
                '.mypy_cache/3.7/ctypes/*',
                '.mypy_cache/3.7/email/*',
                '.mypy_cache/3.7/html/*',
                '.mypy_cache/3.7/importlib/*',
                '.mypy_cache/3.7/importlib_metadata/*',
                '.mypy_cache/3.7/iniconfig/*',
                '.mypy_cache/3.7/jinja2/*',
                '.mypy_cache/3.7/json/*',
                '.mypy_cache/3.7/logging/*',
                '.mypy_cache/3.7/makefun/*',
                '.mypy_cache/3.7/markupsafe/*',
                '.mypy_cache/3.7/multiprocessing/*',
                '.mypy_cache/3.7/os/*',
                '.mypy_cache/3.7/packaging/*',
                '.mypy_cache/3.7/pkg_resources/*',
                '.mypy_cache/3.7/prompt_toolkit/*',
                '.mypy_cache/3.7/prompt_toolkit/application/*',
                '.mypy_cache/3.7/prompt_toolkit/clipboard/*',
                '.mypy_cache/3.7/prompt_toolkit/completion/*',
                '.mypy_cache/3.7/prompt_toolkit/eventloop/*',
                '.mypy_cache/3.7/prompt_toolkit/filters/*',
                '.mypy_cache/3.7/prompt_toolkit/formatted_text/*',
                '.mypy_cache/3.7/prompt_toolkit/input/*',
                '.mypy_cache/3.7/prompt_toolkit/key_binding/*',
                '.mypy_cache/3.7/prompt_toolkit/key_binding/bindings/*',
                '.mypy_cache/3.7/prompt_toolkit/layout/*',
                '.mypy_cache/3.7/prompt_toolkit/lexers/*',
                '.mypy_cache/3.7/prompt_toolkit/output/*',
                '.mypy_cache/3.7/prompt_toolkit/shortcuts/*',
                '.mypy_cache/3.7/prompt_toolkit/shortcuts/progress_bar/*',
                '.mypy_cache/3.7/prompt_toolkit/styles/*',
                '.mypy_cache/3.7/prompt_toolkit/widgets/*',
                '.mypy_cache/3.7/py/*',
                '.mypy_cache/3.7/pyexpat/*',
                '.mypy_cache/3.7/pytest/*',
                '.mypy_cache/3.7/pytest_mock/*',
                '.mypy_cache/3.7/questionary/*',
                '.mypy_cache/3.7/questionary/prompts/*',
                '.mypy_cache/3.7/templateco/*',
                '.mypy_cache/3.7/unittest/*',
                '.mypy_cache/3.7/urllib/*',
                '.mypy_cache/3.7/xml/*',
                '.mypy_cache/3.7/xml/dom/*',
                '.mypy_cache/3.7/xml/parsers/*',
                '.mypy_cache/3.7/xml/parsers/expat/*',
                '.mypy_cache/3.7/xml/sax/*',
                'dist/*']}

install_requires = \
['Jinja2>=3.0,<4.0',
 'colorama>=0.4.4,<0.5.0',
 'importlib-metadata',
 'questionary>=1.10,<2.0',
 'typing-extensions>=4.0.1,<5.0.0']

setup_kwargs = {
    'name': 'templateco',
    'version': '0.1.3',
    'description': '',
    'long_description': '# Templateco\nAn ecosystem for generating customisable templated folders.\n\n## Getting Started\n```shell\n$ pip install templateco templateco-cli templateco-name-plugin\n# Successfully installed!\n\n$ templateco make template folder\n```\n\n1. This will install Templateco, the CLI and the terraform module.\n2. Create a Terraform folder called foo.\n',
    'author': 'Mike Gregory',
    'author_email': 'mike.ja.gregory@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
from pathlib import Path
from typing import Any, Dict, List

from jinja2 import StrictUndefined, Template

logger = logging.getLogger(__name__)


def _template_file(
    input_file: Path,
    template_directory: Path,
    output_directory: Path,
    data: Dict[str, Any],
) -> None:
    """
    Take an input file, and generate a fully templated output file to the filesystem.

    :param input_file: File within the plugin.
    :param output_directory: Directory to output to.
    :param template_directory: Path to the directory to template.
    :param data: Data to template with.
    :return: None.
    """
    relative_path = output_directory / input_file.relative_to(template_directory)

    logger.debug(f"[{relative_path}] Iterating over file.")
    stripped_path = str(relative_path).rstrip(".jinja2")

    templated_path = Template(stripped_path, undefined=StrictUndefined).render(**data)
    logger.debug(f"[{relative_path}] Output name: [{templated_path}].")

    # Make the parent directory if it doesn't exist.
    Path(templated_path).parent.mkdir(parents=True, exist_ok=True)

    with open(input_file.absolute(), "r") as file:
        logger.debug(f"[{relative_path}] Outputting templated file.")
        Template(file.read(), undefined=StrictUndefined).stream(**data).dump(
            templated_path
        )


def generate_output(
    input_directory: Path,
    output_directory: Path,
    folders_to_template: List[str],
    data: Dict[str, Any],
) -> None:
    """
    Output a templated folder structure.

    :param input_directory: Path to input folder structure.
    :param output_directory: Path to output templated files to.
    :param folders_to_template: List of template folders.
    :param data: Dictionary of data.
    :return: None.
    """
    files_to_template = (
        (input_file, input_directory / folder)
        for folder in folders_to_template
        for input_file in (input_directory / folder).rglob("*.jinja2")
    )

    # Iterate over the files to template, and generate them.
    for (input_file, template_directory) in files_to_template:
        _template_file(input_file, template_directory, output_directory, data)
